
/* IPXNET.CPP IPX network library to create and control network games */

#include <mem.h>
#include <string.h>
#include <dos.h>
#include <conio.h>

#include "ipxnode.h"
#include "ipxnet.h"

  #define NET_UPDATE_DELAY 5

  #define NET_TIMEOUT_SCAN_WORKGROUPS 10 // maximum number of seconds to wait for
                                         // work group broadcast data
  #define NET_TIMEOUT_JOIN_WORKGROUP 100

  #define NET_RETRY_SCAN_WORKGROUPS 4 // scan will retry

  #define NET_DATAT_NOWARANT 0  // broadcast workgroup data
  #define NET_DATAT_SEQCOM 1           // net commands data
  #define NET_DATAT_USER 2          // user data

  #define NET_DEFAULT_NAME_MEMBER "Anonymous"
  #define NET_DEFAULT_NAME_SESSION "'s workgroup"

  #define NET_COM_GETINFO 10
  #define NET_COM_LOGIN 20   // slaves commands
  #define NET_COM_LOGINRESULT 30  // master will send this command to a slave
  #define NET_COM_LOGINNEW 31

  #define NET_COM_NEWWORKGROUP 110
//  #define NET_COM_UPDATEWORKGROUP 111
  #define NET_COM_DELETEWORKGROUP 112

  #define NET_COM_LOGIN_ACCESSDENIED 41
  #define NET_COM_MISCINFO 50
  #define NET_COM_MASTERLOGOUT 60  // master/slave will can send this command
  #define NET_COM_SLAVELOGOUT 70
  #define NET_COM_SLAVEKILL 80
  #define NET_COM_STARTUP 100

  typedef struct
  {
    TNetUser   Sender;
    char       Com;
    char       Result;
    char       Data [NODE_DATAFRAGSIZE-sizeof (TNetUser)-2];
  } TNetCommand;

  typedef struct
  {
    char Login;
    char Status;
    char MaxUsers;
  } TMiscInfo;

char NET_installed;
int UsersNode [NET_MAX_USERS]; // node numbers (local machine only)
char NetPassword [NET_MAXLEN_PASSWORD+1];

TNetWorkGroup NetLocalWorkGroup;                // local net work group info
TNetWorkGroup NetWorkGroup [NET_MAX_WORKGROUPS]; // all external work groups info


int NET_GetWorkGroupsCount ()
{
  int j;
  int i;

  j = 0;

  for (i = 0; i < NET_MAX_WORKGROUPS; i++)
  {
    if (strlen (NetWorkGroup [i].Desc))
      j++;
  }
  return j;
}


void NET_Done (void)
{
  if (NET_installed)
  {
    NET_Logout ();
    NET_installed = 0;
  }
  NODE_Done ();
}

// filtered function NODE_Send (can be used by user for delivering
// both user and work group data

int NET_Send (int User, void *Data, int DataSize)
{
  if (NET_installed)
  {
    // call the node library
    if ((NetLocalWorkGroup.Status != NET_STATUS_NOTHING) && (NetLocalWorkGroup.Users [User].Status == NET_USER_STATUS_LOGGED) &&
       (NetLocalWorkGroup.User != User))
    {
      return NODE_Send (UsersNode [User], Data, DataSize, NET_DATAT_USER);
    }
    return 0;
  }
  return 0;
}

int NET_findUser (char *user)
{
  int i;

  for (i = 0; i < NetLocalWorkGroup.UsersCount; i++)
  {
    if (strcmp (NetLocalWorkGroup.Users [i].Name, user) == 0)
    {
      return i;
    }
  }
  return -1;
}

int NET_Send2User (char *user, void *Data, int DataSize)
{
  int User;

  if (NET_installed)
  {
    User = NET_findUser (user);
    if (User == -1)
      return 0;

    // call the node library
    if ((NetLocalWorkGroup.Status != NET_STATUS_NOTHING) && (NetLocalWorkGroup.Users [User].Status == NET_USER_STATUS_LOGGED) &&
       (NetLocalWorkGroup.User != User))
    {
      return NODE_Send (UsersNode [User], Data, DataSize, NET_DATAT_USER);
    }
    return 0;
  }
  return 0;
}

int NET_Receive (int User, void *Data, int DataSize)
{
  if (NET_installed)
  {
    if (NetLocalWorkGroup.Status != NET_STATUS_NOTHING)
      if (NetLocalWorkGroup.Users [User].Status == NET_USER_STATUS_LOGGED)
        if (NetLocalWorkGroup.User != User)
        {
          return NODE_Receive (UsersNode [User], Data, DataSize, NET_DATAT_USER);
        }
    return 0;
  }
  return 0;
}


/*
int NET_SendWorkGroup (int NodeNumber, TNetWorkGroup *Group)
{
  return NODE_Send (NodeNumber, (void *) Group, sizeof (TNetWorkGroup), NET_DATAT_GROUP);
}

int NET_GetWorkGroup (int NodeNumber, TNetWorkGroup *Group)
{
  return NODE_Receive (NodeNumber, (void *) Group, sizeof (TNetWorkGroup), NET_DATAT_GROUP);
}
*/

int NET_SendCom (int NodeNumber, TNetCommand *Com)
{
  // fill the address of the sender
  if (NET_installed)
  {
    Com->Sender.Addr = *NODE_GetLocalAddress ();
    return NODE_Send (NodeNumber, (void *) Com, sizeof (TNetCommand), NET_DATAT_NOWARANT);
  }
  return 0;
}

int NET_SendComToAddr (TIPXAddress *Addr, TNetCommand *Com)
{
  int i, j;

  i = NODE_New (Addr);
  if (i != -1)
  {
     j = NODE_Send (i, (void *) Com, sizeof (TNetCommand), NET_DATAT_SEQCOM);
     NODE_Delete (i);
     return j;
  }
  return 0;
}

int NET_GetCom (int NodeNumber, TNetCommand *Com)
{
  return NODE_Receive (NodeNumber, (void *) Com, sizeof (TNetCommand), NET_DATAT_NOWARANT);
}

/*
int NET_GetBroadcastCom (TNetCommand *Com)
{
  return NODE_Receive (NODE_BROADCAST, (void *) Com, sizeof (TNetCommand), NET_DATAT_BROADCASTCOM);
}


int NET_SendBroadcastCom (TNetCommand *Com)
{
  Com->Sender.Addr = *NODE_GetLocalAddress ();
  return NODE_Send (NODE_BROADCAST, (void *) Com, sizeof (TNetCommand), NET_DATAT_BROADCASTCOM);
}
*/

int NET_SendSeqCom (int NodeNumber, TNetCommand *Com)
{
  Com->Sender.Addr = *NODE_GetLocalAddress ();
  return NODE_Send (NodeNumber, Com, sizeof (TNetCommand), NET_DATAT_SEQCOM);
}

int NET_GetSeqCom (int NodeNumber, TNetCommand *Com)
{
  Com->Sender.Addr = *NODE_GetLocalAddress ();
  return NODE_Receive (NodeNumber, Com, sizeof (TNetCommand), NET_DATAT_SEQCOM);
}

int NET_SendComUser (int User, TNetCommand *Com)
{
  return NET_SendSeqCom (UsersNode [User], Com);
}

int NET_GetComUser (int User, TNetCommand *Com)
{
  return NET_GetSeqCom (UsersNode [User], Com);
}

void NET_SendComEveryone (TNetCommand *Com)
{
  int i;

  for (i = 0; i < NET_MAX_USERS; i++)
  {
    if (NetLocalWorkGroup.Users [i].Status != NET_STATUS_NOTHING)
    {
      if (i != NetLocalWorkGroup.User)
      {
	NET_SendComUser (i, Com);
      }
    }
  }
}

int NET_GetComEveryone (TNetCommand *Com)
{
  int i;

  for (i = 0; i < NET_MAX_USERS; i++)
  {
    if (NetLocalWorkGroup.Users [i].Status != NET_STATUS_NOTHING)
    {
      if ( NET_GetComUser (i, Com) )
      {
        return i;
      }
    }
  }
  return -1;
}



int NET_SendEveryone (void *Data, int DataSize)
{
  int i , j = 0;

  for (i = 0; i < NET_MAX_USERS; i++)
  {
    if (NET_Send (i, Data, DataSize))
    {
      j++;
    }
  }
  return j;
}

int NET_ReceiveEveryone (void *Data, int DataSize)
{
  int i;

  for (i = 0; i < NET_MAX_USERS; i++)
  {
    if ( NET_Receive (i, Data, DataSize) )
    {
      return i;
    }
  }
  return -1;
}

void NET_FlushBroadcastComs (void)
{
  TNetCommand Com;

  while (NET_GetCom (NODE_BROADCAST, &Com))
    ;
}


int InsertWorkGroup (TNetWorkGroup *Group)
{
  int i,j = -1;

  // check if it is our work group
  if ( (Group->Status != NET_STATUS_NOTHING) &&
     ((Group == &NetLocalWorkGroup) ||
      (NetLocalWorkGroup.Status == NET_STATUS_NOTHING) ||
     (memcmp ((void *) &NetLocalWorkGroup.Users [0].Addr,
             (void *) &Group->Users [0].Addr, sizeof (TIPXAddress)))))
  {
    // check if the work group is new
    for (i = 0; i < NET_MAX_WORKGROUPS; i++)
    {
      if ((strlen (NetWorkGroup [i].Desc) == 0) && (j == -1))
      {
        j = i;
      }
      // scan the received work group info
      // check the master address
      if (memcmp ((void *) &NetWorkGroup [i]. Users [0]. Addr,
                  (void *) &Group->Users [0].Addr, sizeof (TIPXAddress)) == 0)
      {
//      if ( (Group->TimeStamp > NetWorkGroup [i].TimeStamp) ||
//           (Group->TimeStamp == 0))
//      {
          NetWorkGroup [i] = *Group;
          return i;
//      }
//      break;
      }
    }
    if (i == NET_MAX_WORKGROUPS)
    {
      // a work group was found, copy data to WorkGroup []
      NetWorkGroup [j] = *Group;
    }
  }
  return j;
}

int DeleteWorkGroup (TNetWorkGroup *Group)
{
  int i;

  // check if the work group is new
  for (i = 0; i < NET_MAX_WORKGROUPS; i++)
  {
    // scan the received work group info
    // check the master address
    if (memcmp ((void *) &NetWorkGroup [i]. Users [0]. Addr, (void *) &Group->Users [0].Addr, sizeof (TIPXAddress)) == 0)
    {
      memset ((void *) &NetWorkGroup [i], 0, sizeof (TNetWorkGroup));
      return i;
    }
  }
  return -1;
}

int UpdateWorkGroups (TNetCommand *Com)
{
  int insert;

  switch (Com->Com)
  {
//    case NET_COM_UPDATEWORKGROUP:
//    {
//        insert =
//      return NET_UPDATE_UPDATEWORKGROUP + InsertWorkGroup ((TNetWorkGroup *)Com->Data);
//    }
    case NET_COM_NEWWORKGROUP:
    {
      insert = InsertWorkGroup ((TNetWorkGroup *)Com->Data);
      if (insert != -1)
      {
        // if insert workgroup fails returns -1
        return NET_UPDATE_NEWWORKGROUP + insert;
      }
    }
    case NET_COM_DELETEWORKGROUP:
    {
      insert = DeleteWorkGroup ((TNetWorkGroup *)Com->Data);
      if (insert != -1)
      {
        // if delete workgroup fails returns -1
        return NET_UPDATE_DELETEWORKGROUP + insert;
      }
    }
  }
  return 0;
}



// scans for all work groups over the net and fills WorkGroup [] data
// 1. BroadCast
void NET_ScanWorkGroups (void)
{
  TNetCommand Com;

  if (NET_installed)
  {
    // start time counting now
    unsigned short t = NODE_GetCounter ();
    int ScanRetry = 0;

  //  NET_FlushBroadcastComs ();

  //  Com.Com = NET_COM_GETINFO;
  //  NET_SendCom (NODE_BROADCAST, &Com);

    // reset all external WorkGroup data
    memset ((void *) &NetWorkGroup, 0, sizeof NetWorkGroup);

    do
    {
      if ((NODE_GetCounter () - t) > NET_TIMEOUT_SCAN_WORKGROUPS)
      {
        t = NODE_GetCounter ();
        ScanRetry++;
      }

      // get command from every user (should be master)
      if (NET_GetCom (NODE_BROADCAST, &Com))
      {
        UpdateWorkGroups (&Com);
      }
    }
    while (ScanRetry < NET_RETRY_SCAN_WORKGROUPS);
    // return total number of work groups
  }
}

int NET_Init (void)
{
  int i;

  if (NET_installed == 0)
  {
    // call node library
    i = NODE_Init ();
    if (i)
    {
      // clear all used variables
      memset ((void *) &NetLocalWorkGroup, 0, sizeof NetLocalWorkGroup);
      memset ((void *) &NetWorkGroup, 0, sizeof NetWorkGroup);

      // set netxxx defaults
      NET_installed = i;

      NET_ScanWorkGroups ();
    }
  }
  return NET_installed;
}



void SendWorkGroupInfo (int Command)
{
  TNetCommand Com;

  Com.Com = Command;
  if (Command == NET_COM_NEWWORKGROUP)
    NetLocalWorkGroup.TimeStamp = 0;
  else
    NetLocalWorkGroup.TimeStamp++;

  (*(TNetWorkGroup *)Com.Data) = NetLocalWorkGroup;

  NET_SendCom (NODE_BROADCAST, &Com);
}



// this updates miscalennous info
void NET_SendMiscInfo ()
{
  TNetCommand Com;
  TMiscInfo Info;

  Com.Com  = NET_COM_MISCINFO;
  Info.MaxUsers = NetLocalWorkGroup.UsersCountMax;
  Info.Status   = NetLocalWorkGroup.Status;
  Info.Login    = NetLocalWorkGroup.Login;
  *((TMiscInfo *) Com.Data) = Info;
  NET_SendComEveryone (&Com);

//  SendWorkGroupInfo (NODE_BROADCAST, NET_COM_NEWWORKGROUP);
}


int NET_EnableLogin (void)
{
  // check if the user is a master
  if ((NetLocalWorkGroup.User == NET_USER_MASTER) &&
      (NetLocalWorkGroup.Status == NET_STATUS_LOGIN))
  {
    // enable login
    NetLocalWorkGroup.Login = NET_LOGIN_ENABLE;
    NET_SendMiscInfo ();
    return 1;
  }
  return 0;
}



int NET_DisableLogin (void)
{
  // check if the user is master
  if ((NetLocalWorkGroup.User == NET_USER_MASTER) &&
      (NetLocalWorkGroup.Status == NET_STATUS_LOGIN))
  {
    // disable login
    NetLocalWorkGroup.Login = NET_LOGIN_DISABLE;
    NET_SendMiscInfo ();
    return 1;
  }
  return 0;
}



int NET_SetMaxUsers (int MaxUsers)
{
  if ((NetLocalWorkGroup.User == NET_USER_MASTER) &&
      (NetLocalWorkGroup.Status == NET_STATUS_LOGIN))
  {
    if ((MaxUsers > 0) && (MaxUsers <= NET_MAX_USERS))
    {
      NetLocalWorkGroup.UsersCountMax = MaxUsers;
      NET_SendMiscInfo ();

      return 1; // success
    }
  }
  return 0; // fail
}




void NET_NODE_NewAll (void)
{
  int i;

  // ALL NODES EXCEPT THE MASTER (ALREADY)
  for (i = 1; i < NET_MAX_USERS; i++)
  {
    // open all neccesary nodes except the master node
    if (NetLocalWorkGroup.Users [i].Status == NET_USER_STATUS_LOGGED)
    {
      UsersNode [i] = NODE_New (&NetLocalWorkGroup.Users [i].Addr);
    }
  }
}

// deletes all nodes except master
void NET_NODE_DeleteAll (void)
{
  int i;

  for (i = 0; i < NET_MAX_USERS; i++)
  {
    // close all nodes except the master node
    if (NetLocalWorkGroup.Users [i].Status == NET_USER_STATUS_LOGGED)
    {
      NODE_Delete (UsersNode [i]);
    }
  }
}




// returns the pointer to workgroup info
TNetWorkGroup *NET_GetWorkGroupInfo (int Number)
{
  // check the number
  if (Number > NET_MAX_WORKGROUPS)
    return NULL; // error, number not in bounds
  return &NetWorkGroup [Number]; // ok
}









void _NET_CopyName (char *Dest, const char *Name)
{
  // adjust the name
  if (strlen (Name) != 0)
  {
    // copy player name to work group 0
    strncpy (Dest, Name, NET_MAXLEN_NAME);
  }
  else
  {
    // copy default member name to work group 0
    strncpy (Dest, NET_DEFAULT_NAME_MEMBER, NET_MAXLEN_NAME);
  }
  Dest [NET_MAXLEN_NAME] = 0;
}

void _NET_CopyDesc (char *Dest, const char *Desc, const char *Name)
{
  // adjust the name
  if (strlen (Desc) != 0)
  {
    // copy player name to work group 0
    strncpy (Dest, Desc, NET_MAXLEN_DESC);
  }
  else
  {
    _NET_CopyName (Dest, Name);
    // copy default member name to work group 0
    strncat (Dest, NET_DEFAULT_NAME_SESSION, NET_MAXLEN_DESC-strlen (Dest));
  }
  Dest [NET_MAXLEN_DESC] = 0;
}

void _NET_CopyPassword (char *Dest, const char *Password)
{
  Dest [NET_MAXLEN_PASSWORD] = 0;
  strncpy (Dest, Password, NET_MAXLEN_PASSWORD);
}


int NET_FindUser (TIPXAddress *Addr)
{
  int i;

  // scan workgroup to check if player is already logged
  for (i = 1; i < NET_MAX_USERS; i++)
  {
    if ((NetLocalWorkGroup.Users [i].Status != NET_USER_STATUS_FREE) &&
       (memcmp ((void *) &NetLocalWorkGroup.Users [i]. Addr, (void *) Addr, sizeof (TIPXAddress)) == 0))
      break;
  }
  if (i == NET_MAX_USERS)
    return 0;
  return i;
}

int NET_LoginUser (TIPXAddress *Addr, const char *Name)
{
  int i;

  // check if already logged
  if (!NET_FindUser (Addr))
  {
    // check the Users count
    if ((NetLocalWorkGroup. UsersCount < NetLocalWorkGroup. UsersCountMax))// &&
        // check if it is enabled to login the user
//      (NetLocalWorkGroup.Login == NET_LOGIN_ENABLE) || (NetLocalWorkGroup.User != NET_USER_MASTER))
    {
      // find the free space to store the member name
      for (i = 1; (i < NetLocalWorkGroup. UsersCountMax) &&
          (NetLocalWorkGroup. Users [i]. Status != NET_USER_STATUS_FREE);
          i++)
        ;

      if (i != NetLocalWorkGroup. UsersCountMax)
      {
        NetLocalWorkGroup. UsersCount++;

        // copy member name
        strcpy (NetLocalWorkGroup. Users [i].Name, Name);
        // copy the node address
        NetLocalWorkGroup.Users [i].Addr = *Addr;
        NetLocalWorkGroup.Users [i].Status = NET_USER_STATUS_LOGGED;

        // open a new node for the member
        UsersNode [i] = NODE_New (Addr);
        return i;
      }
    }
  }
  return 0;
}

int NET_LogoutUser (TIPXAddress *Addr)
{
  int i;

  // check if user is really present in work group data
  i = NET_FindUser (Addr);
  if (i)
  {
    // clear all the user data
    memset ((void *) &NetLocalWorkGroup. Users [i], 0, sizeof (TNetUser));
    NODE_Delete (UsersNode [i]);
    // decrement the number of the Users
    NetLocalWorkGroup. UsersCount--;
    return i;
  }
  return 0;
}





// create a new workgroup
int NET_Create (const char *GameDesc, const char *PlayerName, const char *Password)
{
  int i,j;
  char dsc [256];

  if (NetLocalWorkGroup.Status == NET_STATUS_NOTHING)
  {
//    NODE_FlushPackets ();

    // clear the Users
    memset ((void *) &NetLocalWorkGroup, 0, sizeof NetLocalWorkGroup);
    memset (UsersNode, 0, sizeof UsersNode);

    // opens a node to connect with the master (with self;  for the master)
    NetLocalWorkGroup.Users [0]. Addr = *NODE_GetLocalAddress ();
    NetLocalWorkGroup.Users [0]. Status = NET_USER_STATUS_LOGGED;

    UsersNode [0] = NODE_New (&NetLocalWorkGroup. Users [0].Addr);

    // rescan work group to check if there is a free Work Group
    if (NET_GetWorkGroupsCount () < NET_MAX_WORKGROUPS)
    {
      _NET_CopyDesc (NetLocalWorkGroup.Desc, GameDesc, PlayerName);

      for (i = 0; i < NET_MAX_WORKGROUPS; i++)
      {
	if ((strcmp (NetWorkGroup [i].Desc, NetLocalWorkGroup.Desc) == 0) &&
	   (NetWorkGroup [i].Status != NET_STATUS_NOTHING))
	{
	  strcpy ((char *) &NetLocalWorkGroup.Desc, "");
	  return NET_LOGINRESULT_GROUPALREADYEXISTS;
	}
      }

      // copy the player name
      _NET_CopyName ((char *) &NetLocalWorkGroup.Users [0]. Name, PlayerName);
      // copy the name of the game


      _NET_CopyPassword (NetPassword, Password);
      if (strlen (NetPassword) == 0)
      {
	NetLocalWorkGroup.Password = NET_PASSWORD_NONE;
      }
      else
      {
	NetLocalWorkGroup.Password = NET_PASSWORD_REQUIRED;
      }

      // set the login mode
      NetLocalWorkGroup.UsersCountMax = NET_MAX_USERS;
      NetLocalWorkGroup.UsersCount = 1;
      NetLocalWorkGroup.User = NET_USER_MASTER;
      NetLocalWorkGroup.Status = NET_STATUS_LOGIN;
      NET_EnableLogin ();

      SendWorkGroupInfo (NET_COM_NEWWORKGROUP);

      return NET_LOGINRESULT_SUCCESS;
    }
    return NET_LOGINRESULT_TOOMANYWORKGROUPS;
  }
  return NET_LOGINRESULT_ALREADYLOGGED;
}

int NET_findWorkgroup (char *desc)
{
  int i;

  for (i = 0; i < NET_MAX_WORKGROUPS; i++)
  {
    if ((strcmp (NetWorkGroup [i].Desc, desc) == 0) &&
     (NetWorkGroup [i].Status != NET_STATUS_NOTHING))
      return i;
  }
  return -1;
}


// join a workgroup
int NET_Join (const char *GameDesc, const char *Name, const char *Password)
{
  int i,j;
  unsigned short t;
  TNetCommand Com;
  TNetWorkGroup Group;
  int GroupNumber = -1;

  if (NET_installed)
  {
    if (NetLocalWorkGroup.Status == NET_STATUS_NOTHING)
    {
      for (i = 0; i < NET_MAX_WORKGROUPS; i++)
      {
	if ((strcmp (NetWorkGroup [i].Desc, GameDesc) == 0) &&
	 (NetWorkGroup [i].Status != NET_STATUS_NOTHING))
	  GroupNumber = i;
      }

      if (GroupNumber != -1)
      {
	if (NetWorkGroup [GroupNumber].Status == NET_STATUS_LOGIN)
	{
	  while (NET_Update ());
	  NODE_FlushPackets ();

	  memset ((void *) &NetLocalWorkGroup, 0, sizeof NetLocalWorkGroup);
	  memset ((void *) &UsersNode, 0, sizeof UsersNode);
	  // opens a node to connect with the master - NET_MASTERNODE
	  //  NetLocalWorkGroup = NetWorkGroup [GroupNumber];

	  //copy self address
	  Com.Com = NET_COM_LOGIN;
	  // copy player name
	  _NET_CopyName ((char *) &Com.Sender.Name, Name);
	  _NET_CopyPassword ((char *) &Com.Data, Password);
	  // send command to the master - he will insert the player into the player list
	  // and send a NET_COM_LOGGED

          UsersNode [NET_USER_MASTER] = NODE_New (&NetWorkGroup [GroupNumber]. Users [0].Addr);
          NET_SendCom (NET_NODE_MASTER, &Com);

          t = NODE_GetCounter ();

	  do
          {
            NET_Update ();
            //NET_FlushBroadcastComs ();
            NET_GetComUser (NET_USER_MASTER, &Com);
          }
          while ( ((NODE_GetCounter () - t) < NET_TIMEOUT_JOIN_WORKGROUP) &&
                   (Com.Com != NET_COM_LOGINRESULT));

          if (Com.Com == NET_COM_LOGINRESULT)
          {
	    switch (Com.Result)
            {
              case NET_LOGINRESULT_SUCCESS:
              {
                // copy out the work group info to temporary group
                NetLocalWorkGroup = *(TNetWorkGroup *) Com.Data;
                // find self
                NetLocalWorkGroup.User = NET_FindUser (NODE_GetLocalAddress ());
		NET_NODE_NewAll ();

                DeleteWorkGroup (&NetWorkGroup [GroupNumber]);
                break;
              }
              default:
              {
                NetLocalWorkGroup.Status = NET_STATUS_NOTHING;
                NODE_Delete (NET_NODE_MASTER);
		break;
              }
            }
            return Com.Result;
          }
	  Com.Com = NET_COM_SLAVELOGOUT;
	  NET_SendComEveryone (&Com);

          return NET_LOGINRESULT_TIMEOUT;
        }
        return NET_LOGINRESULT_NOTLOGINMODE;
      }
      return NET_LOGINRESULT_INVALIDGROUP;
    }
    return NET_LOGINRESULT_ALREADYLOGGED; // timeout or NetWorkGroup already full
  }
  return NET_LOGINRESULT_NONETWORK;
}

// join a workgroup
int NET_JoinAddress (TIPXAddress *address, const char *Name, const char *Password)
{
  int i,j;
  unsigned short t = NODE_GetCounter ();
  TNetCommand Com;
  TNetWorkGroup Group;

  if (NET_installed)
  {
    if (NetLocalWorkGroup.Status == NET_STATUS_NOTHING)
    {
      while (NET_Update ());
      NODE_FlushPackets ();

      memset ((void *) &NetLocalWorkGroup, 0, sizeof NetLocalWorkGroup);
      memset ((void *) &UsersNode, 0, sizeof UsersNode);
      // opens a node to connect with the master - NET_MASTERNODE
      //  NetLocalWorkGroup = NetWorkGroup [GroupNumber];

      //copy self address
      Com.Com = NET_COM_LOGIN;
      // copy player name
      _NET_CopyName ((char *) &Com.Sender.Name, Name);
      _NET_CopyPassword ((char *) &Com.Data, Password);
      // send command to the master - he will insert the player into the player list
      // and send a NET_COM_LOGGED

      UsersNode [NET_USER_MASTER] = NODE_New (address);
      NET_SendCom (NET_NODE_MASTER, &Com);

      do
      {
        NET_Update ();
        //NET_FlushBroadcastComs ();
        NET_GetComUser (NET_USER_MASTER, &Com);
      }
      while ( ((NODE_GetCounter () - t) < NET_TIMEOUT_JOIN_WORKGROUP) &&
               (Com.Com != NET_COM_LOGINRESULT));

      if (Com.Com == NET_COM_LOGINRESULT)
      {
        switch (Com.Result)
        {
          case NET_LOGINRESULT_SUCCESS:
          {
            // copy out the work group info to temporary group
            NetLocalWorkGroup = *(TNetWorkGroup *) Com.Data;
            // find self
            NetLocalWorkGroup.User = NET_FindUser (NODE_GetLocalAddress ());
            NET_NODE_NewAll ();
            break;
          }

          case NET_LOGINRESULT_LOGINDISABLED:
          case NET_LOGINRESULT_TOOMANYUSERS:
          case NET_LOGINRESULT_ACCESSDENIED:
          {
            NODE_Delete (NET_NODE_MASTER);
            break;
          }
        }
        return Com.Result;
      }
      return NET_LOGINRESULT_TIMEOUT;
    }
    return NET_LOGINRESULT_ALREADYLOGGED; // timeout or NetWorkGroup already full
  }
  return NET_LOGINRESULT_NONETWORK;
}


unsigned short NetUpdateTime;

int Update_Master (void)
{
  TNetCommand Com;
  int i;
  unsigned short t;

  NODE_ClearWastePackets ();

  // check for a joiner or slaves - broadcasted commands
  while (NET_GetCom (NODE_BROADCAST, &Com))
  {
    i=UpdateWorkGroups (&Com);
    if (i)
      return i;

    switch (Com.Com)
    {
      case NET_COM_GETINFO:
      {
        if (NET_FindUser (&Com.Sender.Addr) == 0)
        {
          i = NODE_New (&Com.Sender.Addr);
//        SendWorkGroupInfo (i, NET_COM_NEWWORKGROUP);
          NODE_Delete (i);
        }
        break;
      }

      case NET_COM_LOGIN:
      {
        Com.Com  = NET_COM_LOGINRESULT;
        // check the password status

        if (NetLocalWorkGroup.Login == NET_LOGIN_ENABLE)
        {
          if (NetLocalWorkGroup.Password == NET_PASSWORD_REQUIRED)
          {
            if (strcmp ((char *) Com.Data, NetPassword))
            {
              Com.Result = NET_LOGINRESULT_ACCESSDENIED;
              NET_SendComToAddr (&Com.Sender.Addr, &Com);
              return NET_UPDATE_LOGINACCESSDENIED;
            }
          }

          for (i = 0; i < NetLocalWorkGroup.UsersCount; i++)
          {
            if (strcmp (NetLocalWorkGroup.Users [i].Name, Com.Sender.Name) == 0)
            {
              break;
            }
          }
          if (i == NetLocalWorkGroup.UsersCount)
          {
            if (NET_FindUser (&Com.Sender.Addr))
            {
              NET_LogoutUser (&Com.Sender.Addr);
//              Com.Result = NET_LOGINRESULT_ALREADYLOGGED;
//              NET_SendComToAddr (&Com.Sender.Addr, &Com);
            }
//            else
            {
              i = NET_LoginUser (&Com.Sender.Addr, Com.Sender.Name);

              if (i)
              {
                // send to all slaves info about new member
                Com.Com = NET_COM_LOGINNEW;
                *((TNetUser *) Com.Data) = NetLocalWorkGroup.Users [i];
                NET_SendComEveryone (&Com);

                // send to joiner he was logged
                Com.Com = NET_COM_LOGINRESULT;
                Com.Result = NET_LOGINRESULT_SUCCESS;
                *(TNetWorkGroup *) Com.Data = NetLocalWorkGroup;
                ((TNetWorkGroup *) Com.Data)->User = i;
                NET_SendComUser (i, &Com);

    //          SendWorkGroupInfo (NODE_BROADCAST, NET_COM_NEWWORKGROUP);

                return NET_UPDATE_LOGIN + i;
              }
              else
              {
                Com.Result = NET_LOGINRESULT_TOOMANYUSERS;
                NET_SendComToAddr (&Com.Sender.Addr, &Com);
              }
            }
            break;
          }
          else
          {
            Com.Result = NET_LOGINRESULT_BADNAME;
            NET_SendComToAddr (&Com.Sender.Addr, &Com);
          }
        }
        else
        {
          Com.Result = NET_LOGINRESULT_LOGINDISABLED;
          NET_SendComToAddr (&Com.Sender.Addr, &Com);
        }
      }
    }
  }

  if (NET_GetComEveryone (&Com))
  {
    switch (Com.Com)
    {

      case NET_COM_SLAVELOGOUT:
      {
       // a slave had post a request to logout
        i = NET_LogoutUser (&Com.Sender.Addr);
        if (i)
        {
          * (TIPXAddress *) Com.Data = Com.Sender.Addr;
          NET_SendComEveryone (&Com);

//        SendWorkGroupInfo (NODE_BROADCAST, NET_COM_NEWWORKGROUP);

          return NET_UPDATE_LOGOUT + i;
        }
      }
    }
  }

  t = NODE_GetCounter ();
  if ((t - NetUpdateTime) > NET_UPDATE_DELAY)
  {
    NetUpdateTime = t;
    SendWorkGroupInfo (NET_COM_NEWWORKGROUP);
  }

  return NET_UPDATE_NOTHING;
}

int Update_Slave (void)
{
  TNetCommand Com;
  TMiscInfo *Info;
  TIPXAddress *Addr;
  TNetUser *User;

  int i;

  while (NET_GetCom (NODE_BROADCAST, &Com))
  {
    i=UpdateWorkGroups (&Com);
    if (i)
      return i;
  }

//  NET_FlushBroadcastComs ();
  // listen commands only from master
  if (NET_GetComUser (NET_USER_MASTER, &Com))
  {
    switch (Com.Com)
    {
      case NET_COM_LOGINNEW:
      {
        User = (TNetUser *) &Com.Data;
        i = NET_LoginUser (&User->Addr, User->Name);
        if (i)
          return NET_UPDATE_LOGIN + i;
        break;
      }

      case NET_COM_SLAVELOGOUT:
      {
       // a slave had post a request to logout
        i = NET_LogoutUser ((TIPXAddress *) Com.Data);

        if (i)
          return NET_UPDATE_LOGOUT + i;
        break;
      }

      case NET_COM_MISCINFO:
      {
        Info = (TMiscInfo *) Com.Data;
        NetLocalWorkGroup.UsersCountMax = Info->MaxUsers;
        NetLocalWorkGroup.Status = Info->Status;
        NetLocalWorkGroup.Login = Info->Login;
        return NET_UPDATE_MISCINFO;
      }

      case NET_COM_MASTERLOGOUT:
      {
        NetLocalWorkGroup.Status = NET_STATUS_NOTHING;
        NET_LogoutUser (NODE_GetLocalAddress ());
        NET_NODE_DeleteAll ();
        return NET_UPDATE_LOGOUT;
      }

      case NET_COM_SLAVEKILL:
      {
        // address of the user to logout is stored in Com.Data
        Addr = (TIPXAddress *) Com.Data;
        if (NET_FindUser (Addr) == NetLocalWorkGroup.User)
        {
          NET_LogoutUser (NODE_GetLocalAddress ());
          InsertWorkGroup (&NetLocalWorkGroup);

          NetLocalWorkGroup.Status = NET_STATUS_NOTHING;
          NET_NODE_DeleteAll ();

          return NET_UPDATE_CLEARED_U;
        }
        else
        {
          i = NET_LogoutUser (Addr);
          if (i)
          {
            return NET_UPDATE_CLEARED + i;
          }
        }

      }
    }
  }
  return NET_UPDATE_NOTHING;
}


int Update_NoOne (void)
{
  int i;
  TNetCommand Com;

  while (NET_GetCom (NODE_BROADCAST, &Com))
  {
    i=UpdateWorkGroups (&Com);
    if (i)
      return i;
  }
  return NET_UPDATE_NOTHING;
}


// call the net to update
int NET_Update (void)
{
  if (NET_installed)
  {
    NODE_ClearWastePackets ();

    switch (NetLocalWorkGroup.Status)
    {
      case NET_STATUS_NOTHING:
      {
        return Update_NoOne ();
      }

      case NET_STATUS_LOGIN:
      case NET_STATUS_STARTUP:
      case NET_STATUS_SESSION:
      {
        // check the net user type
        switch (NetLocalWorkGroup.User)
        {
          // code for the master computer
          case NET_USER_MASTER:
          {
            return Update_Master ();
          }

          // code for the slave computer
          default:
          {
            return Update_Slave ();
          }
        }
      }
    }
    return NET_UPDATE_NOTHING;
  }
  return NET_UPDATE_NONETWORK;
}


int NET_Synchronize (void)
{
  TNetCommand Com;
  int i;
  int j = 1; //number of synchronized computers (except this station)

  if (NetLocalWorkGroup.Status == NET_STATUS_STARTUP)
  {
    switch (NetLocalWorkGroup.User)
    {
      case NET_USER_MASTER:
      {
        // wait for other peers
        while (j < NetLocalWorkGroup.UsersCount)
        {
          if (NET_GetComEveryone (&Com) != -1)
          {
            if (Com.Com == NET_COM_STARTUP)
            {
              j++;
            }
          }
        }
        Com.Com = NET_COM_STARTUP;
        NET_SendComEveryone (&Com);
        break;
      }

      default:
      {
        Com.Com = NET_COM_STARTUP;
        NET_SendComUser (NET_USER_MASTER, &Com);

        // wait for the master
        while ( (!NET_GetComUser (NET_USER_MASTER, &Com)) ||
               (Com.Com != NET_COM_STARTUP));
        break;
      }
    }

    NetLocalWorkGroup.Status = NET_STATUS_SESSION;
    return 1;
  }
  return 0;
}


// this function is available only for the master computer
int NET_Startup (void)
{
  int i;
//  unsigned short t;
  TNetCommand Com;

  if ((NetLocalWorkGroup.User == NET_USER_MASTER) &&
      (NetLocalWorkGroup.Status == NET_STATUS_LOGIN))
  {
    // set status to running mode
    NetLocalWorkGroup.Status = NET_STATUS_STARTUP;
    NetLocalWorkGroup.Login = NET_LOGIN_DISABLE;
    NetLocalWorkGroup.UsersCountMax = NetLocalWorkGroup.UsersCount;

    NET_SendMiscInfo ();
    return 1; // ok
  }
  else
   return 0; // user is not a master
}



int NET_Logout (void)
{
  int i;
  TNetCommand Com;

  // set status to logged mode
  if (NetLocalWorkGroup.Status != NET_STATUS_NOTHING)
  {
    if (NET_installed)
    {
      switch (NetLocalWorkGroup.User)
      {

        case NET_USER_MASTER:
        {
          Com.Com = NET_COM_MASTERLOGOUT;
          NET_SendComEveryone (&Com);

	  NODE_Delay (10);

	  SendWorkGroupInfo (NET_COM_DELETEWORKGROUP);
	  break;
	}

	default:
	{
	  Com.Com = NET_COM_SLAVELOGOUT;
	  NET_SendComEveryone (&Com);

	  NET_LogoutUser (NODE_GetLocalAddress ());
	  InsertWorkGroup (&NetLocalWorkGroup);
	  break;
	}
      }
    }

    // delete *all* opened nodes (except local)
    NetLocalWorkGroup.Status = NET_STATUS_NOTHING;

    NET_NODE_DeleteAll ();
    return 1;
  }
  return 0;
}


int NET_ClearSlave (int User)
{
  TNetCommand Com;

  if ((NetLocalWorkGroup.User == NET_USER_MASTER) &&
      (NetLocalWorkGroup.Status != NET_STATUS_NOTHING))
  {
    if ((User < NET_MAX_USERS) && (User) &&
       (NetLocalWorkGroup.Users [User].Status == NET_USER_STATUS_LOGGED))
    {
      Com.Com = NET_COM_SLAVEKILL;
      // copy the address of the slave to the data
      memcpy (Com.Data, (void *) &NetLocalWorkGroup. Users [User].Addr, sizeof (TIPXAddress));
      // send the command to clear the slave
      NET_SendComEveryone (&Com);
      // clear the user from the master list
      NET_LogoutUser (&NetLocalWorkGroup.Users [User]. Addr);

//      SendWorkGroupInfo (NODE_BROADCAST, NET_COM_NEWWORKGROUP);
      return 1;
    }
  }
  return 0;
}

char *NET_WhatsMyName ()
{
  if (NetLocalWorkGroup.Status != NET_STATUS_NOTHING)
  {
    return NetLocalWorkGroup.Users [NetLocalWorkGroup.User].Name;
  }
  return NULL;
}
