
#ifndef _WIN_H_
  #define _WIN_H_

  #ifdef __cplusplus
  extern "C" {
  #endif

  #include <stddef.h>
  #include "csf.h"

  #define MAX_SUBWIN 256
  #define WIN_ITEMLEN 100

  #define W_CENTER -1

  #define BLACK 0
  #define BLUE 1
  #define GREEN 2
  #define CYAN 3
  #define RED 4
  #define MAGENTA 5
  #define BROWN 6
  #define LIGHTGRAY 7
  #define DARKGRAY 8
  #define LIGHTBLUE 9
  #define LIGHTGREEN 10
  #define LIGHTCYAN 11
  #define LIGHTRED 12
  #define LIGHTMAGENTA 13
  #define YELLOW 14
  #define WHITE 15

  #define B_BLACK (0)
  #define B_BLUE (1 * 16)
  #define B_GREEN (2 * 16)
  #define B_CYAN (3 * 16)
  #define B_RED (4 * 16)
  #define B_MAGENTA (5 * 16)
  #define B_BROWN (6 * 16)
  #define B_LIGHTGRAY (7 * 16)
  #define B_DARKGRAY (8 * 16)
  #define B_LIGHTBLUE (9 * 16)
  #define B_LIGHTGREEN (10 * 16)
  #define B_LIGHTCYAN (11 * 16)
  #define B_LIGHTRED (12 * 16)
  #define B_LIGHTMAGENTA (13 * 16)
  #define B_YELLOW (14 * 16)
  #define B_WHITE (15 * 16)

  #define B_BLINK 128


  #define WIN_COM_ESC   0x0001
  #define WIN_COM_ENTER 0x0002
  #define WIN_COM_DEL   0x0004
  #define WIN_COM_INS   0x0010
//  #define WIN_COM_TAB   0x0020
  #define WIN_COM_F1    0x0040
  #define WIN_COM_EXIT  -2
  #define WIN_COM_INPUT -1

  #define WIN_FRAME 1
  #define WIN_NONEFRAME 0

  #define WIN_setPassiveFrame(win) WIN_setFrame(win,'','','','','','')
  #define WIN_setActiveFrame(win) WIN_setFrame(win,'','','','','','')
  #define WIN_setNoneFrame(win) WIN_setFrame(win, 0, 0, 0, 0, 0, 0)

  #define ITEM_SIZE 256
  #define ITEM_SIZE_DESC 80

  #define WIN_ITEMT_MENU 1
  #define WIN_ITEMT_LIST 8
  #define WIN_ITEMT_BUTTON 5
  #define WIN_ITEMT_COMMENT 2
  #define WIN_ITEMT_INPUT 3
  #define WIN_ITEMT_INPUT2 4
  #define WIN_ITEMT_SCROLLBAR 6
  #define WIN_ITEMT_CHECKBOX 7

  typedef struct {
    unsigned int esc   : 1;
    unsigned int enter : 1;
    unsigned int del   : 1;
    unsigned int ins   : 1;
//    unsigned int tab   : 1;
    unsigned int f10   : 1;
  } TWINCom;

  typedef struct {
    int type;
    char desc [ITEM_SIZE_DESC];
    char item [ITEM_SIZE];
    int maxLen;
    int curPos;
    int step;
    int pageStep;
    int value;
    int scrollBar;
    int scrollBarSelect;
    long  cursorType;
    char *miniStatus;
  } TWINItem;

  typedef struct {
    int frameActive;
    int framePassive;
    int frameSelect;
    int content;
    int selectPassive;
    int item;
    int select;
    int highlight;
    int h_barSelect;
  } TWINColor;

  typedef struct TWIN_ {
    char title [WIN_ITEMLEN];    // pointer to title
    int frame;         // the Win has frame ?
    int shadow;
    int scrollBar;
    int scrollBarSelect;
    int buttonPressed;
    int selectable;
    int x;
    int y;
    int sizeX;
    int sizeY;
    int csizeX;
    int csizeY;
    struct TWIN_ *owner;
    struct TWIN_ *subWin [MAX_SUBWIN];
//    int enabled;
    struct TWIN_ *humbleWin;
    int humbleItem;
    int subWinActive;

    TWINColor color;

    int contentStart;
    int contentCur;     // current content
    int contentCount;   // maximum number of items
    int contentMax;
    TWINItem *content;  // items

    TWINCom  commsAvail;// available commands
    char **miniHelp;     // pointer to minihelp string
    int curMiniHelp;
    char *videoBuf;     // pointer to videobuffer
    char *saveBuf;      // pointer to self save
    long videoBufSize;   // size of videobuffer

    void **sprSavePtr;  // pointer to bitmap save buffer for graphics modes
    void **sprPtr;
    char *sprBuf; // pointer to allocated sprites in WIN_SetBackground
    char frameChar [6]; // used frame characters
    char BGChar;
  } TWIN;


  extern void WIN_delete ( TWIN *win);
  extern int WIN_insertItem ( TWIN *win, int Item, int Type,
                              char *Desc, char *Str, int Value,
                              char *MiniStatus, int maxLen);

//  extern int WIN_insertItem ( TWIN *win, int Item = -1,
//                    int Type = WIN_ITEMT_COMMENT,
//                    char *Desc = NULL, char *Str = NULL, int Value = -1,
//                    char *MiniStatus = NULL, int maxLen = -1);


  extern int WIN_newItem ( TWIN *win, int Item, int Type,
                   char *Desc, char *Str, int Value,
                   char *MiniStatus, int maxLen);

//  extern int WIN_newItem ( TWIN *win, int Item, int Type = WIN_ITEMT_COMMENT,
//                 char *Desc = NULL, char *Str = NULL, int Value = 0,
//                 char *MiniStatus = NULL, int maxLen = -1);

  extern int WIN_changeItem ( TWIN *win, int Item, char *ItemDesc);
  extern void WIN_deleteItem ( TWIN *win, int Item);
  extern void WIN_deleteContent ( TWIN *win);
  extern void WIN_setBounds ( TWIN *win, int x, int y, int size_x, int size_y);
  extern void WIN_scrollBarEnable ( TWIN *win);
  extern void WIN_scrollBarDisable ( TWIN *win);
  extern void WIN_setColors ( TWIN *win, int Content_C,
                      int Select_C, int Highlight_C,
                      int Item_C, int PassiveSelect_C,
                      int ActiveFrame_C, int PassiveFrame_C);
//  extern void WIN_setColors ( TWIN *win, int Content_C,
//                    int Select_C = -1, int Highlight_C = -1,
//                    int Item_C = -1, int PassiveSelect_C = -1,
//                    int ActiveFrame_C = -1, int PassiveFrame_C = -1);

  extern int WIN_appendItem (TWIN *win, int itemType, char *desc, char *itemStr,
                      int value, char *miniStatus, int maxLen);
//  extern int WIN_appendItem (TWIN *win, int itemType = WIN_ITEMT_COMMENT, char *desc = NULL, char *itemStr = NULL,
//                    int value = -1, char *miniStatus = NULL, int maxLen = -1);

  extern int WIN_appendButton (TWIN *win, char *desc, char *itemStr, int value,
                        char *miniStatus, int maxLen);
//  extern int WIN_appendButton (TWIN *win, char *desc, char *itemStr, int value = 0,
//                      char *miniStatus = NULL, int maxLen = -1);

  extern int WIN_appendMenu (TWIN *win, char *desc, char *itemStr, int value,
                      char *miniStatus, int maxLen);
//  extern int WIN_appendMenu (TWIN *win, char *desc, char *itemStr, int value = 0,
//                    char *miniStatus = NULL, int maxLen = -1);

  extern int WIN_appendList (TWIN *win, char *desc, char *itemStr, int value,
                      char *miniStatus, int maxLen);
//  extern int WIN_appendList (TWIN *win, char *desc, char *itemStr, int value = 0,
//                    char *miniStatus = NULL, int maxLen = -1);

  extern int WIN_appendCheckBox (TWIN *win, char *desc, char *itemStr, int value,
                          char *miniStatus, int maxLen);
//  extern int WIN_appendCheckBox (TWIN *win, char *desc=NULL, char *itemStr=NULL, int value=-1,
//                        char *miniStatus=NULL, int maxLen=-1);

  extern int WIN_appendComment (TWIN *win, char *itemStr); // = NULL


  extern void WIN_setTitle ( TWIN *win, const char *Title);
  extern void WIN_redrawScreen (void);
  extern void WIN_cls ( TWIN *win);
  extern int WIN_init (void);
  extern void WIN_done (void);
  extern void WIN_setFrame ( TWIN *win, char C0, char C1, char C2, char C3, char C4, char C5);
  extern void WIN_puts ( TWIN *win, int X, int Y, const char *str, int Color);
  extern void WIN_redrawFrame ( TWIN *win);
  extern void WIN_printMiniStatus (char *Desc);
  extern void WIN_hide ( TWIN *win);
  extern void WIN_show ( TWIN *win);
  extern int WIN_printContent ( TWIN *win, int Start); // = 0
  extern void WIN_hide ( TWIN *win);
  extern void WIN_show ( TWIN *win);
  extern void WIN_close ( TWIN *win);
  extern void WIN_printMiniStatus (char *Desc);
  extern void WIN_printMiniHelp ( TWIN *win);
  extern int WIN_open ( TWIN *win);
  extern void WIN_openActive ( TWIN *win);
  extern void WIN_openPassive ( TWIN *win);
  extern void WIN_openNoneFrame ( TWIN *win);
  extern void WIN_orintMiniStatus (char *Desc);
  extern void WIN_scrollUp ( TWIN *win);
  extern void WIN_scrollDn ( TWIN *win);
  extern void WIN_scrollHome ( TWIN *win);
  extern void WIN_scrollEnd ( TWIN *win);
  extern void WIN_scrollPageUp ( TWIN *win);
  extern void WIN_scrollPageDn ( TWIN *win);
  extern int WIN_insert ( TWIN *win);
  extern int WIN_exec ( TWIN *win);
  extern void WIN_setMiniStatusWin ( TWIN *win);
  extern void WIN_setMiniHelpString ( TWIN *win, char **Str);
  extern void WIN_setMiniHelpWin ( TWIN *win);
  extern void WIN_setDefaultMiniHelp (char **DefMiniHelp);
  extern  TWIN *WIN_new (int ContentMax, int Frame); // = 0, = WIN_FRAME
  extern int WIN_newViewer ( TWIN *win, char *FileName, char *MiniStatus); // = NULL
  extern void WIN_setWaitHandler ( int (*Handler) (void));
  extern int WIN_changeItemType (TWIN *win, int Item, int Type);
  extern void WIN_moreMiniHelp (TWIN *win);
  extern void WIN_closeAll (void);
  extern void _SetCursorPos (int X, int Y);
//  extern void WIN_setBG (TWIN *win, char Char);
  extern void WIN_setSpr (TWIN *win, TCSF *spr); // = NULL
  extern void WIN_setDesktop (char backGround, char color, TCSF *spr); // = NULL
//  extern void WIN_getItem (char *str, TWIN *win, int content);
  extern int WIN_getValue (TWIN *win, int content);
  extern TWINItem *WIN_item (TWIN *win, int item);
  extern void WIN_setHumble (TWIN *win, TWIN *humbleWin, int humbleItem);
  extern long WIN_value (TWIN *win);
  extern char *WIN_itemStr (TWIN *win);
  extern TWIN *WIN_newWinButton (char *desc, int value);
  extern int WIN_insertSubWin (TWIN *win, TWIN *subWin);
  extern int WIN_deleteSubWin (TWIN *win, TWIN *subWin);
  extern TWIN *WIN_subWin (TWIN *win);
  extern void WIN_updateHumble (TWIN *win, int enter);
  extern char WIN_waitKey (TWIN *win);
  extern int WIN_waitHandler (void);
  extern int WIN_strlen (const char *str);
  extern void WIN_nextSubWin (TWIN *win);
  extern void WIN_previousSubWin (TWIN *win);

  #ifdef __cplusplus
  }
  #endif

#endif
